/*******************************************************************************
 * Copyright (C) Altimetrik 2018. All rights reserved.
 * 
 * This software is the confidential and proprietary information
 * of Altimetrik. You shall not disclose such Confidential Information
 * and shall use it only in accordance with the terms and conditions
 * entered into with Altimetrik.
 ******************************************************************************/
package com.altimetrik.playground.innovation.controller;

import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.validation.BindingResult;
import org.springframework.validation.FieldError;
import org.springframework.validation.ObjectError;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.ResponseStatus;
import org.springframework.web.context.request.WebRequest;
import org.springframework.web.servlet.mvc.method.annotation.ResponseEntityExceptionHandler;

import com.altimetrik.playground.innovation.bean.BaseResponse;
import com.altimetrik.playground.innovation.bean.ResponseModel;
import com.altimetrik.playground.innovation.bean.StatusDetails;
import com.altimetrik.playground.innovation.exception.PgApplicationException;

public abstract class BaseController extends ResponseEntityExceptionHandler {

	private static final String MESSAGE_DESCRIPTION = "messageDescription";

	private static final String ERROR_CODE = "errorCode";

	protected static final String PLATFORM_UNHANDLED_EXCEPTION_THROWN = "PG unhandled exception thrown.";

	protected static final String PLATFORM_FATAL_EXCEPTION_THROWN = "PG fatal exception thrown.";

	protected static final String PLATFORM_SYSTEM_EXCEPTION_THROWN = "PG system exception thrown.";

	protected static final String PLATFORM_VALIDATION_ERROR = "Validation errors encountered";

	protected static final String PLATFORM_APPLICATION_EXCEPTION_THROWN = "PG application exception thrown.";

	@ExceptionHandler(PgApplicationException.class)
	public ResponseEntity<BaseResponse> resolveApplicationException(HttpServletRequest request, HttpServletResponse response, Exception exception) {

		PgApplicationException appException = (PgApplicationException) exception;
		StatusDetails errorStatus = getErrorStatus(appException.getMessage(),
				appException.getLocalizedMessage(appException.getMessage(), new Locale(""), appException.getArguments()));
		BaseResponse errorResponse = new BaseResponse(errorStatus);
		return new ResponseEntity<>(errorResponse, HttpStatus.OK);
	}

	@Override
	@ResponseStatus(HttpStatus.BAD_REQUEST)
	protected ResponseEntity<Object> handleMethodArgumentNotValid(MethodArgumentNotValidException ex, HttpHeaders headers, HttpStatus status, WebRequest request) {
		final List<FieldError> fieldErrors = ex.getBindingResult().getFieldErrors();
		Map<String, Set<String>> errorsMap = fieldErrors.stream()
				.collect(Collectors.groupingBy(FieldError::getField, Collectors.mapping(FieldError::getDefaultMessage, Collectors.toSet())));
		return new ResponseEntity(errorsMap.isEmpty() ? ex : errorsMap, headers, status);
	}

	@ExceptionHandler(Exception.class)
	public ResponseEntity<BaseResponse> resolveUnhandledException(HttpServletRequest request, HttpServletResponse response, Exception exception) {
		exception.printStackTrace();
		StatusDetails errorStatus = getErrorStatus(PLATFORM_UNHANDLED_EXCEPTION_THROWN, exception.getMessage());
		BaseResponse errorResponse = new BaseResponse(errorStatus);
		return new ResponseEntity<>(errorResponse, HttpStatus.INTERNAL_SERVER_ERROR);
	}

	/**
	 * Creates exception status object with the necessary information.
	 * @param errorCode - error code from exception class.
	 * @param errorMessage - error message.
	 * @return - status object.
	 */
	protected StatusDetails getErrorStatus(String errorCode, String errorMessage) {
		StatusDetails errorResponse = new StatusDetails();
		errorResponse.setErrorCode(errorCode);
		errorResponse.setStatusCode(StatusDetails.FAILED);
		errorResponse.setMessageDescription(errorMessage);
		return errorResponse;
	}

	/**
	 * Prepares validation error response from binding result.
	 * @param bindingResult - which holds the validation errors.
	 * @param responseModel - response model.
	 */
	protected void prepareValidationErrorResponse(BindingResult bindingResult, ResponseModel responseModel) {

		List<Object> errorArguments = new ArrayList<>();
		StringBuilder messageDescription = new StringBuilder();
		StringBuilder errorCodes = new StringBuilder();

		StatusDetails status = new StatusDetails();
		status.setStatusCode(StatusDetails.FAILED);

		bindingResult.getFieldErrors().forEach(error -> {
			if (messageDescription.length() > 0) {
				messageDescription.append("; ");
			}
			messageDescription.append("@" + error.getField() + ": " + error.getDefaultMessage()).toString();
			if (errorCodes.length() > 0) {
				errorCodes.append("; ");
			}
			errorCodes.append(error.getCodes()[0]);
			errorArguments.add(error.getRejectedValue());
		});

		status.setErrorCode(errorCodes.toString());
		status.setMessageDescription(messageDescription.toString());
		responseModel.setStatus(status);
	}

	/**
	 * Filters the errors from the errorList for the given objectname.
	 * @param errorsList - List of ObjectErrors
	 * @param objectName - Name of the model object
	 * @return list of errors.
	 */
	protected List getErrorsForObject(List<?> errorsList, String objectName) {
		List filteredErrors = new ArrayList<ObjectError>();
		for (Object object : errorsList) {
			if (objectName.equals(((ObjectError) object).getObjectName())) {
				filteredErrors.add(object);
			}
		}
		return filteredErrors;
	}

	@SuppressWarnings("unchecked")
	public static <T> T getNativeException(Throwable exp, Class<T> requiredType) {
		if (requiredType != null) {
			if (requiredType.isInstance(exp)) {
				return (T) exp;
			}
			else if (exp instanceof Exception) {
				return getNativeException(((Exception) exp).getCause(), requiredType);
			}
		}
		return null;
	}

}
